////  CHIRPS Rainy Season Data Stack:

/*
MIT License

Copyright 2020 Forrest R. Stevens

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

////	Imports:
var polygon = /* color: #d63000 */ee.Geometry.Polygon(
        [[[13.103586077484692, -28.592835709475807],
          [39.7123637362572, -27.01859619099294],
          [44.0094074581217, 0.3468648106177058],
          [6.9916721653796685, 0.1180468261018772]]]),
    image_collection = ee.ImageCollection("UCSB-CHG/CHIRPS/PENTAD"),
    buffers = ee.FeatureCollection("users/Forrest/KAZAVA/HH_buff1km");

////  Study area polygon:
//var myfeatures = ee.FeatureCollection("ft:12eapYEwGyFqcxzju2yzy82w3VJPKDZsHG0nng1Bz");

////  We could alternatively specify this from points here:
//var polygon = ee.Geometry.Rectangle(-123.06, 44.47, -122.96, 44.60);

////  Or as another alternative, draw a polygon using the interface
////    and include it as a variable in the above imports section.



///////////////////////////////////////////////////////////////////////////////
////////  BEGIN:  Aggregated functions for mapping and reduction:

////  Create year, day filter and a simple aggregate for estimating beginning 
////    and ending of seasonality:
var aggregate_beginning = function(year){
  var start_year = year;
  var end_year = year;
  if (end_day_beg < start_day_beg) {
    end_year = end_year + 1;
  }
	//var filtered_collection = ee.ImageCollection(image_collection)
	var filtered_collection = ee.ImageCollection(data_to_process)
    .filterDate(new ee.Date.parse('YYYY DDD', start_year + ' ' + start_day_beg), new ee.Date.parse('YYYY DDD', end_year + ' ' + end_day_beg))
    //.filter(ee.Filter.calendarRange(start_day, end_day))
    .filterBounds(polygon)
    .map(mask_day)
    .select('day_mask')
    .min()
    .subtract( ee.Date.parse('YYYY DDD', start_year + ' ' + start_day_beg).millis())
    .divide(86400000)  //  Convert from milliseconds to days
    .clip(polygon);
  
  var agg = ee.Image(filtered_collection).clip(polygon);
	
  return agg;
};

var aggregate_ending = function(year){
  var start_year = year;
  var end_year = year;
  if (end_day_end < start_day_end) {
    end_year = end_year + 1;
  }
	//var filtered_collection = ee.ImageCollection(image_collection)
	var filtered_collection = ee.ImageCollection(data_to_process)
    .filterDate(new ee.Date.parse('YYYY DDD', start_year + ' ' + start_day_end), new ee.Date.parse('YYYY DDD', end_year + ' ' + end_day_end))
    //.filter(ee.Filter.calendarRange(start_day, end_day))
    .filterBounds(polygon)
    .map(mask_day)
    .select('day_mask')
    .max()
    .subtract( ee.Date.parse('YYYY DDD', start_year + ' ' + start_day_end).millis())
    .divide(86400000)  //  Convert from milliseconds to days
    .clip(polygon);
  
  var agg = ee.Image(filtered_collection).clip(polygon);
	
  return agg;
};


////  Create index with 1/1/year time stamp for beginning:
var create_index_w_time_stamp_beginning = function(year){
	var agg = aggregate_beginning(year);
	var time_start = new Date(year,1,1);
	var time_end = new Date(year + 1, 1,1);
	
	agg = ee.Image(agg).setMulti({'year':year, 'system:time_start':time_start.valueOf(), 'system:time_end':time_end.valueOf()});
	
	return agg;
};

////  Create index with 1/1/year time stamp for ending:
var create_index_w_time_stamp_ending = function(year){
	var agg = aggregate_ending(year);
	var time_start = new Date(year,1,1);
	var time_end = new Date(year + 1, 1,1);
	
	agg = ee.Image(agg).setMulti({'year':year, 'system:time_start':time_start.valueOf(), 'system:time_end':time_end.valueOf()});
	
	return agg;
};

//  Use this function to add bands for NDVI, time, etc. to imagery:
var add_bands = function(image) {
  image = ee.Image(image);
  return image
    // Add an NDVI band:
    //.addBands(image.normalizedDifference(['sur_refl_b02', 'sur_refl_b01']).rename('NDVI')).float()
    // Add a milliseconds-since-1971 band:
    .addBands(ee.Image.constant(ee.Number.parse(image.date().millis())).rename('day').float());
};

//  Use this function to add mask and add day layer:
var mask_day = function(image) {
  var day = image.select('day').clip(polygon);
  var day_mask = day.updateMask(image.select(layer).gt(threshold));
  return image.addBands(day_mask.rename('day_mask'));
};

////  Use this function to calculate growing season:
////    NOTE: this could be accomplished using per-band match, but is less
////      flexible because we are left with a multi-band image not a collection.
//var growing_seasons = data_ending.toBands().add( ee.Number(parseInt(start_day_end, 10) + (365 - parseInt(start_day_beg, 10))) ).subtract( data_beginning.toBands() );
var aggregate_growing_seasons = function(year){
  var beginning = ee.ImageCollection(data_beginning).filterMetadata('year', 'equals', year);
  var ending = ee.ImageCollection(data_ending).filterMetadata('year', 'equals', year);
  
  var agg = ending.first().add( ee.Number(parseInt(start_day_end, 10) + (365 - parseInt(start_day_beg, 10))) ).subtract( beginning.first() );
	
	var time_start = new Date(year,1,1);
	var time_end = new Date(year + 1, 1,1);
	
	agg = ee.Image(agg).setMulti({'year':year, 'system:time_start':time_start.valueOf(), 'system:time_end':time_end.valueOf()});
	
  return agg;
};

///////////////////////////////////////////////////////////////////////////////
////////  BEGIN:  General configuration parameters:

////  Years to process:
var start_year = 1981;
var end_year = 2018;
var years = [];
for (var i = start_year; i <= end_year; i++) {
  years.push(i);
}

////  Select layer from each image and the threshold to apply for selecting
////    the date:
var layer = 'precipitation';
var threshold = 10;



///////////////////////////////////////////////////////////////////////////////
////////  BEGIN:  Process data for final data products:

var data_to_process = image_collection.map(add_bands);

////  Create estimates of beginning of wet-season:
var start_day_beg = '275';    //  Oct 1
var end_day_beg = '031';      //  Jan 30
var data_beginning = ee.ImageCollection(years.map(create_index_w_time_stamp_beginning));
print(data_beginning);

var data_october = data_beginning.map(function (x) { return (ee.Image(x).lt(ee.Image(31.0))) }).reduce(ee.Reducer.sum());

////  Create estimates of ending of wet-season:
var start_day_end = '031';    //  Jan 30
var end_day_end = '181';      //  Jun 30
var data_ending = ee.ImageCollection(years.map(create_index_w_time_stamp_ending));
print(data_ending);

var trends_beginning = data_beginning.formaTrend();

var cv_beginning = data_beginning
  .reduce(ee.Reducer.stdDev())
  .divide( data_beginning.reduce(ee.Reducer.mean() ) );

var cv_ending = data_ending
  .reduce(ee.Reducer.stdDev())
  .divide( data_ending.reduce(ee.Reducer.mean() ) );


var growing_seasons = ee.ImageCollection(years.map(aggregate_growing_seasons));
print(growing_seasons);
var cv_growing_seasons = growing_seasons
  .reduce(ee.Reducer.stdDev())
  .divide( growing_seasons.reduce(ee.Reducer.mean() ) );
var trends_growing_seasons = growing_seasons.formaTrend();
print(trends_growing_seasons);

///////////////////////////////////////////////////////////////////////////////
////////  BEGIN:  Visualization of output:
var SAMpalette = ['00ff00','008000', '808000', 'ffff00','ffA500', 'ff0000','800000','8c2a04'];
SAMpalette.reverse();

Map.addLayer(data_beginning, {min: 0.0, max: 100.0, palette:SAMpalette}, 'Beginning Day');
Map.addLayer(data_october, {min: 0.0, max: 30.0, palette:SAMpalette}, 'Beginning Day in October');
Map.addLayer(data_ending, {min: 0.0, max: 100.0, palette:SAMpalette}, 'Ending Day');
Map.addLayer(growing_seasons, {min: 50.0, max: 300.0, palette:SAMpalette}, 'Growing Season Days');
Map.addLayer(data_beginning.filter(ee.Filter.eq('year',2000)), {min: 0, max: 100.0, palette:SAMpalette}, 'Precipitation Day - 2000');
Map.addLayer(cv_beginning, {min: -0.5, max: 1.0, palette:SAMpalette}, 'Beginning CV');
Map.addLayer(trends_beginning, {bands:'long-trend', min: -1.0, max: 1.0, palette:SAMpalette}, 'Beginning Trend');
Map.addLayer(cv_growing_seasons, {min: -0.5, max: 0.5, palette:SAMpalette}, 'Season CV');
Map.addLayer(trends_growing_seasons, {bands:'long-trend', min: -1.0, max: 1.0, palette:SAMpalette}, 'Season Trend');

////  Add trend, masked by statistical significance:
var trends_beginning_sig = trends_beginning.select('long-tstat').abs().gt(1.96).multiply( trends_beginning.select('long-tstat') );
Map.addLayer(trends_beginning_sig, {min: -10.0, max: 10.0, palette:SAMpalette}, 'Beginning Trend, signif.');

////  Add trend, masked by statistical significance:
var trends_growing_seasons_sig = trends_growing_seasons.select('long-tstat').abs().gt(1.96).multiply( trends_growing_seasons.select('long-tstat') );
Map.addLayer(trends_growing_seasons_sig, {min: -10.0, max: 10.0, palette:SAMpalette}, 'Season Trend, signif.');

Map.addLayer(buffers, {}, 'Buffers', true, 0.3);

//Map.setCenter(-123.05, 44.53, 11);
//Map.centerObject(polygon, 11);
Map.centerObject(buffers);

////  Further diagnostics:
//print(trends);

//var sfTimeSeries =
//  Chart.image.series(data_beginning, polygon, ee.Reducer.mean(), 200);

////  Visualize time series in the Console:
//print(sfTimeSeries, 'LineChart');



///////////////////////////////////////////////////////////////////////////////
////////  BEGIN:  Export of final products:

var mean_growing_season = growing_seasons.mean();
Map.addLayer(mean_growing_season, {min: 50.0, max: 300.0, palette:SAMpalette}, 'Mean Growing Season Days');

Export.image.toDrive(
  {
      image:ee.Image(trends_growing_seasons), 
      description:'TrendsGrowingSeasonExport', 
      folder:'KAZAVA', 
      crs:'EPSG:4326', 
      region:polygon.toGeoJSON(), 
      scale:6000
  }
);

////  Conduct zonal aggregation by buffers:
// Add reducer output to the Features in the collection.
var reducer = ee.Reducer.mean().combine(ee.Reducer.minMax(), null, true).combine(ee.Reducer.stdDev(), null, true);

var buffersReduced = mean_growing_season.reduceRegions({
  collection: buffers,
  reducer: reducer,
  scale: 30,
});

// Print the first feature, to illustrate the result.
print(ee.Feature(buffersReduced.first()));

Export.table.toDrive(buffersReduced, 'ExportFeatures', '', '', 'SHP');


////  Conduct zonal sampling by buffers:
////  Stack our image collection, stacking together the individual images
////    into a single image with multiple bands.  Last, use .select() to rename 
////    the bands:
var years = ee.List(years);
var multiband = ee.Image(growing_seasons.first());
multiband = years.slice(1,years.size()).iterate(function(year,collection) {
  return ee.Image(collection).addBands({srcImg:ee.Image(growing_seasons.filterMetadata('year', 'equals', ee.Number(year)).first()), overwrite:false});
}, multiband);
multiband = ee.Image(multiband).select(ee.Image(multiband).bandNames(), years.map(function(year) {return ee.String('year_').cat(ee.String(year))}));
print(multiband);

////  Note the mapped function to convert the buffered features
////    to centroid points, but we're sampling at 1000 m:
var buffersSampled = multiband.sampleRegions({
  collection: buffers.map(function (x) { return ee.Feature(x).centroid() } ),
  scale: 1000,
});

// Print the first feature, to illustrate the result.
print(ee.Feature(buffersSampled.first()));

Export.table.toDrive(buffersSampled, 'ExportSampledFeatures', '', '', 'CSV');
